@@ -13,6 +13,7 @@
 import pandas as pd
 import numpy as np
 from tqdm import tqdm
+from collections import OrderedDict
 from hover import module_config
 from hover.core import Loggable
 from hover.utils.bokeh_helper import auto_label_color
@@ -27,8 +28,11 @@ from bokeh.models import (
 )
 from .local_config import (
     dataset_help_widget,
+    dataset_default_sel_table_columns,
+    dataset_default_sel_table_kwargs,
     COLOR_GLYPH_TEMPLATE,
     DATASET_SUBSET_FIELD,
+    embedding_field,
 )
 
 
@@ -55,7 +59,25 @@ class SupervisableDataset(Loggable):
 
     FEATURE_KEY = "feature"
 
-    def __init__(
+    def __init__(self, *args, **kwargs):
+        """
+        ???+ note "Set up data subsets, widgets, and supplementary data structures."
+
+            See `self.setup_dfs` for parameter details.
+        """
+        self._info("Initializing...")
+        self.setup_dfs(*args, **kwargs)
+        self.df_deduplicate()
+        self.compute_feature_index()
+        self.setup_widgets()
+        # self.setup_label_coding() # redundant if setup_pop_table() immediately calls this again
+        self.setup_file_export()
+        self.setup_pop_table(width_policy="fit", height_policy="fit")
+        self.setup_sel_table(width_policy="fit", height_policy="fit")
+        self._vectorizer_lookup = OrderedDict()
+        self._good(f"{self.__class__.__name__}: finished initialization.")
+
+    def setup_dfs(
         self,
         raw_dictl,
         train_dictl=None,
@@ -65,7 +87,8 @@ class SupervisableDataset(Loggable):
         label_key="label",
     ):
         """
-        ???+ note "Create (1) dictl and df forms and (2) the mapping between categorical and string labels."
+        ???+ note "Subroutine of the constructor that creates standard-format DataFrames."
+
             | Param         | Type   | Description                          |
             | :------------ | :----- | :----------------------------------- |
             | `raw_dictl`   | `list` | list of dicts holding the **to-be-supervised** raw data |
@@ -75,7 +98,6 @@ class SupervisableDataset(Loggable):
             | `feature_key` | `str`  | the key for the feature in each piece of data |
             | `label_key`   | `str`  | the key for the `**str**` label in supervised data |
         """
-        self._info("Initializing...")
 
         def dictl_transform(dictl, labels=True):
             """
@@ -106,50 +128,76 @@ class SupervisableDataset(Loggable):
 
             return [burner(_d) for _d in dictl]
 
-        self.dictls = {
+        # standardize records
+        dictls = {
             "raw": dictl_transform(raw_dictl, labels=False),
             "train": dictl_transform(train_dictl),
             "dev": dictl_transform(dev_dictl),
             "test": dictl_transform(test_dictl),
         }
 
-        self.synchronize_dictl_to_df()
-        self.df_deduplicate()
-        self.synchronize_df_to_dictl()
-        self.setup_widgets()
-        # self.setup_label_coding() # redundant if setup_pop_table() immediately calls this again
-        self.setup_file_export()
-        self.setup_pop_table(width_policy="fit", height_policy="fit")
-        self.setup_sel_table(width_policy="fit", height_policy="fit")
-        self._good(f"{self.__class__.__name__}: finished initialization.")
+        # initialize dataframes
+        self.dfs = dict()
+        for _key, _dictl in dictls.items():
+            if _dictl:
+                _df = pd.DataFrame(_dictl)
+                assert self.__class__.FEATURE_KEY in _df.columns
+                assert "label" in _df.columns
+            else:
+                _df = pd.DataFrame(columns=[self.__class__.FEATURE_KEY, "label"])
 
-    def copy(self, use_df=True):
+            self.dfs[_key] = _df
+
+    def copy(self):
         """
         ???+ note "Create another instance, copying over the data entries."
-            | Param    | Type   | Description                          |
-            | :------- | :----- | :----------------------------------- |
-            | `use_df` | `bool` | whether to use the df or dictl form  |
-        """
-        if use_df:
-            self.synchronize_df_to_dictl()
-        return self.__class__(
-            raw_dictl=self.dictls["raw"],
-            train_dictl=self.dictls["train"],
-            dev_dictl=self.dictls["dev"],
-            test_dictl=self.dictls["test"],
-            feature_key=self.__class__.FEATURE_KEY,
-            label_key="label",
-        )
 
-    def to_pandas(self, use_df=True):
+            Also copy data structures that don't get created in the new instance.
+        """
+        dataset = self.__class__.from_pandas(self.to_pandas())
+        dataset._vectorizer_lookup.update(self._vectorizer_lookup)
+        return dataset
+
+    def compute_feature_index(self):
+        """
+        ???+ note "Allow lookup by feature value without setting it as the index."
+
+            Assumes that feature values are unique. The reason not to just set the feature as the index is because integer indices work smoothly with Bokeh `DataSource`s, NumPy `array`s, and Torch `Tensor`s.
+        """
+        feature_to_subset_idx = {}
+        for _subset, _df in self.dfs.items():
+            _values = _df[self.__class__.FEATURE_KEY].values
+            for i, _val in enumerate(_values):
+                if _val in feature_to_subset_idx:
+                    raise ValueError(
+                        f"Expected unique feature values, found duplicate {_val}"
+                    )
+                feature_to_subset_idx[_val] = (_subset, i)
+        self.feature_to_subset_idx = feature_to_subset_idx
+
+    def locate_by_feature_value(self, value, auto_recompute=True):
+        """
+        ???+ note "Find the subset and index given a feature value."
+
+            Assumes that the value is present and detects if the subset and index found is consistent with the value.
+        """
+        subset, index = self.feature_to_subset_idx[value]
+
+        current_value = self.dfs[subset].at[index, self.__class__.FEATURE_KEY]
+        if current_value != value:
+            if auto_recompute:
+                self._warn("locate_by_feature_value mismatch. Recomputing index.")
+                self.compute_feature_index()
+                # if ever need to recompute twice, there must be a bug
+                return self.locate_by_feature_value(value, auto_recompute=False)
+            else:
+                raise ValueError("locate_by_feature_value mismatch.")
+        return subset, index
+
+    def to_pandas(self):
         """
         ???+ note "Export to a pandas DataFrame."
-            | Param    | Type   | Description                          |
-            | :------- | :----- | :----------------------------------- |
-            | `use_df` | `bool` | whether to use the df or dictl form  |
         """
-        if not use_df:
-            self.synchronize_dictl_to_df()
         dfs = []
         for _subset in ["raw", "train", "dev", "test"]:
             _df = self.dfs[_subset].copy()
@@ -166,11 +214,18 @@ class SupervisableDataset(Loggable):
             | :------- | :----- | :----------------------------------- |
             | `df` | `DataFrame` | with a "SUBSET" field dividing subsets |
         """
+        SUBSETS = cls.SCRATCH_SUBSETS + cls.PUBLIC_SUBSETS + cls.PRIVATE_SUBSETS
+
+        if DATASET_SUBSET_FIELD not in df.columns:
+            raise ValueError(
+                f"Expecting column '{DATASET_SUBSET_FIELD}' in the DataFrame which takes values from {SUBSETS}"
+            )
+
         dictls = {}
         for _subset in ["raw", "train", "dev", "test"]:
-            dictls[_subset] = df[df[DATASET_SUBSET_FIELD] == _subset].to_dict(
-                orient="records"
-            )
+            _sub_df = df[df[DATASET_SUBSET_FIELD] == _subset]
+            dictls[_subset] = _sub_df.to_dict(orient="records")
+
         return cls(
             raw_dictl=dictls["raw"],
             train_dictl=dictls["train"],
@@ -182,6 +237,15 @@ class SupervisableDataset(Loggable):
     def setup_widgets(self):
         """
         ???+ note "Create `bokeh` widgets for interactive data management."
+
+            Operations:
+            -   PUSH: push updated dataframes to linked `explorer`s.
+            -   COMMIT: added selected points to a specific subset `dataframe`.
+            -   DEDUP: cross-deduplicate across all subset `dataframe`s.
+            -   VIEW: view selected points of linked `explorer`s.
+                -   the link can be different from that for PUSH. Typically all the `explorer`s sync their selections, and only an `annotator` is linked to the `dataset`.
+            -   PATCH: update a few edited rows from VIEW result to the dataset.
+            -   EVICT: remove a few rows from both VIEW result and linked `explorer` selection.
         """
         self.update_pusher = Button(
             label="Push", button_type="success", height_policy="fit", width_policy="min"
@@ -205,34 +269,65 @@ class SupervisableDataset(Loggable):
             height_policy="fit",
             width_policy="min",
         )
+        self.selection_patcher = Button(
+            label="Update Row Values",
+            button_type="warning",
+            height_policy="fit",
+            width_policy="min",
+        )
+        self.selection_evictor = Button(
+            label="Evict Rows from Selection",
+            button_type="primary",
+            height_policy="fit",
+            width_policy="min",
+        )
 
         def commit_base_callback():
             """
             COMMIT creates cross-duplicates between subsets.
+            Changes dataset rows.
+            No change to explorers.
 
             - PUSH shall be blocked until DEDUP is executed.
+            - PATCH shall be blocked until PUSH is executed.
+            - EVICT shall be blocked until PUSH is executed.
             """
             self.dedup_trigger.disabled = False
             self.update_pusher.disabled = True
+            self.selection_patcher.disabled = True
+            self.selection_evictor.disabled = True
 
         def dedup_base_callback():
             """
             DEDUP re-creates dfs with different indices than before.
+            Changes dataset rows.
+            No change to explorers.
 
             - COMMIT shall be blocked until PUSH is executed.
+            - PATCH shall be blocked until PUSH is executed.
+            - EVICT shall be blocked until PUSH is executed.
             """
             self.update_pusher.disabled = False
             self.data_committer.disabled = True
+            self.selection_patcher.disabled = True
+            self.selection_evictor.disabled = True
             self.df_deduplicate()
 
         def push_base_callback():
             """
             PUSH enforces df consistency with all linked explorers.
+            No change to dataset rows.
+            Changes explorers.
 
             - DEDUP could be blocked because it stays trivial until COMMIT is executed.
             """
             self.data_committer.disabled = False
             self.dedup_trigger.disabled = True
+            # empty the selection table, then allow PATCH and EVICT
+            self.sel_table.source.data = dict()
+            self.sel_table.source.selected.indices = []
+            self.selection_patcher.disabled = False
+            self.selection_evictor.disabled = False
 
         self.update_pusher.on_click(push_base_callback)
         self.data_committer.on_click(commit_base_callback)
@@ -249,14 +344,20 @@ class SupervisableDataset(Loggable):
 
         return column(
             self.help_div,
+            # population table and directly associated widgets
             row(
                 self.update_pusher,
                 self.data_committer,
                 self.dedup_trigger,
-                self.selection_viewer,
                 self.file_exporter,
             ),
             self.pop_table,
+            # selection table and directly associated widgets
+            row(
+                self.selection_viewer,
+                self.selection_patcher,
+                self.selection_evictor,
+            ),
             self.sel_table,
         )
 
@@ -270,10 +371,8 @@ class SupervisableDataset(Loggable):
 
             Note: the reason we need this is due to `self.dfs[key] = ...`-like assignments. If DF operations were all in-place, then the explorers could directly access the updates through their `self.dfs` references.
         """
-        # local import to avoid import cycles
-        from hover.core.explorer.base import BokehBaseExplorer
 
-        assert isinstance(explorer, BokehBaseExplorer)
+        explorer.link_dataset(self)
 
         def callback_push():
             df_dict = {_v: self.dfs[_k] for _k, _v in subset_mapping.items()}
@@ -294,6 +393,8 @@ class SupervisableDataset(Loggable):
             | `subset_mapping` | `dict` | `dataset` -> `explorer` subset mapping |
         """
 
+        explorer.link_dataset(self)
+
         def callback_commit(event):
             for sub_k, sub_v in subset_mapping.items():
                 sub_to = event.item
@@ -304,12 +405,6 @@ class SupervisableDataset(Loggable):
                     )
                     return
 
-                # take selected slice, ignoring ABSTAIN'ed rows
-                # CAUTION: applying selected_idx from explorer.source to self.df
-                #     this assumes that the source and the df have consistent entries.
-                # Consider this:
-                #    keep_cols = self.dfs[sub_k].columns
-                #    sel_slice = explorer.dfs[sub_v].iloc[selected_idx][keep_cols]
                 sel_slice = self.dfs[sub_k].iloc[selected_idx]
                 valid_slice = sel_slice[
                     sel_slice["label"] != module_config.ABSTAIN_DECODED
@@ -352,21 +447,48 @@ class SupervisableDataset(Loggable):
             isinstance(subsets, list) and len(subsets) > 0
         ), "Expected a non-empty list of subsets"
 
+        explorer.link_dataset(self)
+
         def callback_view():
             sel_slices = []
             for subset in subsets:
-                selected_idx = explorer.sources[subset].selected.indices
+                selected_idx = sorted(explorer.sources[subset].selected.indices)
                 sub_slice = explorer.dfs[subset].iloc[selected_idx]
                 sel_slices.append(sub_slice)
 
             selected = pd.concat(sel_slices, axis=0)
-
-            # replace this with an actual display (and analysis)
             self._callback_update_selection(selected)
 
+        def callback_evict():
+            # create sets for fast index discarding
+            subset_to_indicies = {}
+            for subset in subsets:
+                indicies = set(explorer.sources[subset].selected.indices)
+                subset_to_indicies[subset] = indicies
+
+            # from datatable index, get feature values to look up dataframe index
+            sel_source = self.sel_table.source
+            raw_indicies = sel_source.selected.indices
+            for i in raw_indicies:
+                feature_value = sel_source.data[self.__class__.FEATURE_KEY][i]
+                subset, idx = self.locate_by_feature_value(feature_value)
+                subset_to_indicies[subset].discard(idx)
+
+            # assign indices back to change actual selection
+            for subset in subsets:
+                indicies = sorted(list(subset_to_indicies[subset]))
+                explorer.sources[subset].selected.indices = indicies
+
+            self._good(
+                f"Selection table: evicted {len(raw_indicies)} points from selection."
+            )
+            # refresh the selection table
+            callback_view()
+
         self.selection_viewer.on_click(callback_view)
+        self.selection_evictor.on_click(callback_evict)
         self._good(
-            f"Subscribed {explorer.__class__.__name__} to selection view: {subsets}"
+            f"Subscribed {explorer.__class__.__name__} to selection table: {subsets}"
         )
 
     def setup_label_coding(self, verbose=True, debug=False):
@@ -412,11 +534,14 @@ class SupervisableDataset(Loggable):
         for _key in [*self.__class__.PUBLIC_SUBSETS, *self.__class__.PRIVATE_SUBSETS]:
             _invalid_indices = None
             assert "label" in self.dfs[_key].columns
-            _mask = self.dfs[_key]["label"].apply(lambda x: x in self.label_encoder)
-            _invalid_indices = np.where(_mask is False)[0].tolist()
+            _mask = self.dfs[_key]["label"].apply(
+                lambda x: int(x in self.label_encoder)
+            )
+            # DO NOT change the "==" to "is"; False in pandas is not False below
+            _invalid_indices = np.where(_mask == 0)[0].tolist()
             if _invalid_indices:
                 self._fail(f"Subset {_key} has invalid labels:")
-                self._print({self.dfs[_key].loc[_invalid_indices]})
+                self._print(self.dfs[_key].loc[_invalid_indices])
                 if raise_exception:
                     raise ValueError("invalid labels")
 
@@ -441,7 +566,7 @@ class SupervisableDataset(Loggable):
                 timestamp = current_time("%Y%m%d%H%M%S")
                 path_root = f"hover-dataset-export-{timestamp}"
 
-            export_df = self.to_pandas(use_df=True)
+            export_df = self.to_pandas()
 
             if export_format == "Excel":
                 export_path = f"{path_root}.xlsx"
@@ -458,7 +583,7 @@ class SupervisableDataset(Loggable):
             else:
                 raise ValueError(f"Unexpected export format {export_format}")
 
-            self._good(f"Saved DataFrame to {export_path}")
+            self._good(f"saved Pandas DataFrame version to {export_path}")
 
         # assign the callback, keeping its reference
         self._callback_export = callback_export
@@ -532,29 +657,44 @@ class SupervisableDataset(Loggable):
             | :--------- | :----- | :--------------------------- |
             | `**kwargs` |        | forwarded to the `DataTable` |
         """
+
         sel_source = ColumnDataSource(dict())
-        sel_columns = [
-            TableColumn(
-                field=_col,
-                title=_col,
-            )
-            for _col in self.dfs["train"].columns
-        ]
-        self.sel_table = DataTable(source=sel_source, columns=sel_columns, **kwargs)
+        sel_columns = dataset_default_sel_table_columns(self.__class__.FEATURE_KEY)
+        table_kwargs = dataset_default_sel_table_kwargs(self.__class__.FEATURE_KEY)
+        table_kwargs.update(kwargs)
+        self.sel_table = DataTable(
+            source=sel_source, columns=sel_columns, **table_kwargs
+        )
 
         def update_selection(selected_df):
             """
-            Callback function.
+            To be triggered as a subroutine of `self.selection_viewer`.
             """
-            # push results to bokeh data source
             sel_source.data = selected_df.to_dict(orient="list")
+            # now that selection table has changed, clear sub-selection
+            sel_source.selected.indices = []
 
             self._good(
-                f"Selection updater: latest selection with {selected_df.shape[0]} entries."
+                f"Selection table: latest selection with {selected_df.shape[0]} entries."
             )
 
         self._callback_update_selection = update_selection
 
+        def patch_edited_selection():
+            sel_source = self.sel_table.source
+            raw_indices = sel_source.selected.indices
+            for i in raw_indices:
+                feature_value = sel_source.data[self.__class__.FEATURE_KEY][i]
+                subset, idx = self.locate_by_feature_value(feature_value)
+                for key in sel_source.data.keys():
+                    self.dfs[subset].at[idx, key] = sel_source.data[key][i]
+
+            self._good(f"Selection table: edited {len(raw_indices)} dataset rows.")
+            # if edited labels (which is common), then population has changed
+            self._callback_update_population()
+
+        self.selection_patcher.on_click(patch_edited_selection)
+
     def df_deduplicate(self):
         """
         ???+ note "Cross-deduplicate data entries by feature between subsets."
@@ -594,61 +734,56 @@ class SupervisableDataset(Loggable):
             after[_key] = self.dfs[_key].shape[0]
             self._info(f"--subset {_key} rows: {before[_key]} -> {after[_key]}.")
 
-    def synchronize_dictl_to_df(self):
-        """
-        ???+ note "Re-make dataframes from lists of dictionaries."
-        """
-        self.dfs = dict()
-        for _key, _dictl in self.dictls.items():
-            if _dictl:
-                _df = pd.DataFrame(_dictl)
-                assert self.__class__.FEATURE_KEY in _df.columns
-                assert "label" in _df.columns
-            else:
-                _df = pd.DataFrame(columns=[self.__class__.FEATURE_KEY, "label"])
+        self.compute_feature_index()
 
-            self.dfs[_key] = _df
+    @property
+    def vectorizer_lookup(self):
+        return self._vectorizer_lookup
 
-    def synchronize_df_to_dictl(self):
-        """
-        ???+ note "Re-make lists of dictionaries from dataframes."
-        """
-        self.dictls = dict()
-        for _key, _df in self.dfs.items():
-            self.dictls[_key] = _df.to_dict(orient="records")
+    @vectorizer_lookup.setter
+    def vectorizer_lookup(self, *args, **kwargs):
+        self._fail("assigning vectorizer lookup by reference is forbidden.")
 
-    def compute_2d_embedding(self, vectorizer, method, **kwargs):
+    def compute_nd_embedding(self, vectorizer, method, dimension=2, **kwargs):
         """
-        ???+ note "Get embeddings in the xy-plane and return the dimensionality reducer."
+        ???+ note "Get embeddings in n-dimensional space and return the dimensionality reducer."
             Reference: [`DimensionalityReducer`](https://github.com/phurwicz/hover/blob/main/hover/core/representation/reduction.py)
 
             | Param        | Type       | Description                        |
             | :----------- | :--------- | :--------------------------------- |
             | `vectorizer` | `callable` | the feature -> vector function     |
             | `method`     | `str`      | arg for `DimensionalityReducer`    |
+            | `dimension`  | `int`      | dimension of output embedding      |
             | `**kwargs`   |            | kwargs for `DimensionalityReducer` |
         """
         from hover.core.representation.reduction import DimensionalityReducer
 
+        # register the vectorizer for scenarios that may need it
+        self.vectorizer_lookup[dimension] = vectorizer
+
         # prepare input vectors to manifold learning
         fit_subset = [*self.__class__.SCRATCH_SUBSETS, *self.__class__.PUBLIC_SUBSETS]
         trans_subset = [*self.__class__.PRIVATE_SUBSETS]
 
         assert not set(fit_subset).intersection(set(trans_subset)), "Unexpected overlap"
+        assert isinstance(dimension, int) and dimension >= 2
+        embedding_cols = [embedding_field(dimension, i) for i in range(dimension)]
 
         # compute vectors and keep track which where to slice the array for fitting
         feature_inp = []
         for _key in fit_subset:
-            feature_inp += self.dfs[_key][self.__class__.FEATURE_KEY].tolist()
+            feature_inp.extend(self.dfs[_key][self.__class__.FEATURE_KEY].tolist())
         fit_num = len(feature_inp)
         for _key in trans_subset:
-            feature_inp += self.dfs[_key][self.__class__.FEATURE_KEY].tolist()
-        trans_arr = np.array([vectorizer(_inp) for _inp in tqdm(feature_inp)])
+            feature_inp.extend(self.dfs[_key][self.__class__.FEATURE_KEY].tolist())
+        trans_arr = np.array(
+            [vectorizer(_inp) for _inp in tqdm(feature_inp, desc="Vectorizing")]
+        )
 
         # initialize and fit manifold learning reducer using specified subarray
         self._info(f"Fit-transforming {method.upper()} on {fit_num} samples...")
         reducer = DimensionalityReducer(trans_arr[:fit_num])
-        fit_embedding = reducer.fit_transform(method, **kwargs)
+        fit_embedding = reducer.fit_transform(method, dimension=dimension, **kwargs)
 
         # compute embedding of the whole dataset
         self._info(
@@ -671,28 +806,47 @@ class SupervisableDataset(Loggable):
                 continue
             for _key in _subset:
                 _length = self.dfs[_key].shape[0]
-                self.dfs[_key]["x"] = pd.Series(
-                    _embedding[start_idx : (start_idx + _length), 0]
-                )
-                self.dfs[_key]["y"] = pd.Series(
-                    _embedding[start_idx : (start_idx + _length), 1]
-                )
+                for _i in range(dimension):
+                    _col = embedding_cols[_i]
+                    self.dfs[_key][_col] = pd.Series(
+                        _embedding[start_idx : (start_idx + _length), _i]
+                    )
                 start_idx += _length
 
+        self._good(f"Computed {dimension}-d embedding in columns {embedding_cols}")
         return reducer
 
-    def loader(self, key, vectorizer, batch_size=64, smoothing_coeff=0.0):
+    def compute_2d_embedding(self, vectorizer, method, **kwargs):
         """
-        ???+ note "Prepare a torch `Dataloader` for training or evaluation."
+        ???+ note "Get embeddings in the xy-plane and return the dimensionality reducer."
+            A special case of `compute_nd_embedding`.
+
             | Param        | Type       | Description                        |
             | :----------- | :--------- | :--------------------------------- |
-            | `key`        | `str`      | subset of data, e.g. `"train"`     |
             | `vectorizer` | `callable` | the feature -> vector function     |
-            | `batch_size` | `int`      | size per batch                     |
-            | `smoothing_coeff` | `float` | portion of probability to equally split between classes |
+            | `method`     | `str`      | arg for `DimensionalityReducer`    |
+            | `**kwargs`   |            | kwargs for `DimensionalityReducer` |
+        """
+        reducer = self.compute_nd_embedding(vectorizer, method, dimension=2, **kwargs)
+        return reducer
+
+    def loader(self, key, *vectorizers, batch_size=64, smoothing_coeff=0.0):
+        """
+        ???+ note "Prepare a torch `Dataloader` for training or evaluation."
+            | Param         | Type          | Description                        |
+            | :------------ | :------------ | :--------------------------------- |
+            | `key`         | `str`         | subset of data, e.g. `"train"`     |
+            | `vectorizers` | `callable`(s) | the feature -> vector function(s)  |
+            | `batch_size`  | `int`         | size per batch                     |
+            | `smoothing_coeff` | `float`   | portion of probability to equally split between classes |
         """
         # lazy import: missing torch should not break the rest of the class
-        from hover.utils.torch_helper import vector_dataloader, one_hot, label_smoothing
+        from hover.utils.torch_helper import (
+            VectorDataset,
+            MultiVectorDataset,
+            one_hot,
+            label_smoothing,
+        )
 
         # take the slice that has a meaningful label
         df = self.dfs[key][self.dfs[key]["label"] != module_config.ABSTAIN_DECODED]
@@ -702,27 +856,62 @@ class SupervisableDataset(Loggable):
             raise ValueError(f"Subset {key} has too few samples ({df.shape[0]})")
         batch_size = min(batch_size, df.shape[0])
 
+        # prepare output vectors
         labels = df["label"].apply(lambda x: self.label_encoder[x]).tolist()
-        features = df[self.__class__.FEATURE_KEY].tolist()
         output_vectors = one_hot(labels, num_classes=len(self.classes))
-
-        self._info(f"Preparing {key} input vectors...")
-        input_vectors = [vectorizer(_f) for _f in tqdm(features)]
         if smoothing_coeff > 0.0:
             output_vectors = label_smoothing(
                 output_vectors, coefficient=smoothing_coeff
             )
+
+        # prepare input vectors
+        assert len(vectorizers) > 0, "Expected at least one vectorizer"
+        multi_flag = len(vectorizers) > 1
+        features = df[self.__class__.FEATURE_KEY].tolist()
+
+        input_vector_lists = []
+        for _vec_func in vectorizers:
+            self._info(f"Preparing {key} input vectors...")
+            _input_vecs = [_vec_func(_f) for _f in tqdm(features, desc="Vectorizing")]
+            input_vector_lists.append(_input_vecs)
+
         self._info(f"Preparing {key} data loader...")
-        loader = vector_dataloader(input_vectors, output_vectors, batch_size=batch_size)
+        if multi_flag:
+            assert len(input_vector_lists) > 1, "Expected multiple lists of vectors"
+            loader = MultiVectorDataset(input_vector_lists, output_vectors).loader(
+                batch_size=batch_size
+            )
+        else:
+            assert len(input_vector_lists) == 1, "Expected only one list of vectors"
+            input_vectors = input_vector_lists[0]
+            loader = VectorDataset(input_vectors, output_vectors).loader(
+                batch_size=batch_size
+            )
         self._good(
-            f"Prepared {key} loader consisting of {len(features)} examples with batch size {batch_size}"
+            f"Prepared {key} loader with {len(features)} examples; {len(vectorizers)} vectors per feature, batch size {batch_size}"
         )
         return loader
 
 
 class SupervisableTextDataset(SupervisableDataset):
     """
-    ???+ note "Can add text-specific methods."
+    ???+ note "`SupervisableDataset` whose primary feature is `text`."
     """
 
     FEATURE_KEY = "text"
+
+
+class SupervisableImageDataset(SupervisableDataset):
+    """
+    ???+ note "`SupervisableDataset` whose primary feature is `image`."
+    """
+
+    FEATURE_KEY = "image"
+
+
+class SupervisableAudioDataset(SupervisableDataset):
+    """
+    ???+ note "`SupervisableDataset` whose primary feature is `audio`."
+    """
+
+    FEATURE_KEY = "audio"
