@@ -1,12 +1,19 @@
 """
 ???+ note "Base class(es) for ALL explorer implementations."
 """
+import pandas as pd
 from abc import ABC, abstractmethod
 from collections import OrderedDict, defaultdict
-from bokeh.plotting import figure
+from bokeh.events import SelectionGeometry
 from bokeh.models import ColumnDataSource, Slider
+from bokeh.plotting import figure
 from hover.core import Loggable
+from hover.core.local_config import (
+    is_embedding_field,
+    blank_callback_on_change as blank,
+)
 from hover.utils.bokeh_helper import bokeh_hover_tooltip
+from hover.utils.meta.traceback import RichTracebackABCMeta
 from hover.utils.misc import RootUnionFind
 from .local_config import SEARCH_SCORE_FIELD
 
@@ -20,13 +27,10 @@ STANDARD_PLOT_TOOLS = [
     "lasso_select",
     # make inspections
     "hover",
-    # navigate changes
-    "undo",
-    "redo",
 ]
 
 
-class BokehBaseExplorer(Loggable, ABC):
+class BokehBaseExplorer(Loggable, ABC, metaclass=RichTracebackABCMeta):
     """
     ???+ note "Base class for visually exploring data with `Bokeh`."
         Assumes:
@@ -43,8 +47,10 @@ class BokehBaseExplorer(Loggable, ABC):
     """
 
     SUBSET_GLYPH_KWARGS = {}
+    DEFAULT_SUBSET_MAPPING = {_k: _k for _k in ["raw", "train", "dev", "test"]}
 
-    MANDATORY_COLUMNS = ["label", "x", "y"]
+    PRIMARY_FEATURE = None
+    MANDATORY_COLUMNS = ["label"]
     TOOLTIP_KWARGS = {"label": True, "coords": True, "index": True}
 
     def __init__(self, df_dict, **kwargs):
@@ -57,9 +63,9 @@ class BokehBaseExplorer(Loggable, ABC):
 
             1. settle the figure settings by using child class defaults & kwargs overrides
             2. settle the glyph settings by using child class defaults
-            3. create widgets that child classes can override
-            4. create data sources the correspond to class-specific data subsets.
-            5. activate builtin search callbacks depending on the child class.
+            3. set up dataframes to sync with
+            4. create widgets that child classes can override
+            5. create data sources the correspond to class-specific data subsets.
             6. initialize a figure under the settings above
         """
         self.figure_kwargs = {
@@ -77,7 +83,6 @@ class BokehBaseExplorer(Loggable, ABC):
         self._setup_dfs(df_dict)
         self._setup_sources()
         self._setup_widgets()
-        self._activate_search_builtin()
 
     @classmethod
     def from_dataset(cls, dataset, subset_mapping, *args, **kwargs):
@@ -95,7 +100,18 @@ class BokehBaseExplorer(Loggable, ABC):
 
         assert isinstance(dataset, SupervisableDataset)
         df_dict = {_v: dataset.dfs[_k] for _k, _v in subset_mapping.items()}
-        return cls(df_dict, *args, **kwargs)
+        explorer = cls(df_dict, *args, **kwargs)
+        explorer.link_dataset(dataset)
+        return explorer
+
+    def link_dataset(self, dataset):
+        """
+        ???+ note "Get tied to a dataset, which is common for explorers."
+        """
+        if not hasattr(self, "linked_dataset"):
+            self.linked_dataset = dataset
+        else:
+            assert self.linked_dataset is dataset, "cannot link to two datasets"
 
     def view(self):
         """
@@ -125,8 +141,11 @@ class BokehBaseExplorer(Loggable, ABC):
         self._info("Setting up widgets")
         self._dynamic_widgets = OrderedDict()
         self._dynamic_callbacks = OrderedDict()
-        self._setup_search_highlight()
+        self._dynamic_resources = OrderedDict()
+        self._setup_search_widgets()
+        self._setup_selection_option()
         self._setup_subset_toggle()
+        self._setup_axes_dropdown()
 
     @abstractmethod
     def _layout_widgets(self):
@@ -137,23 +156,39 @@ class BokehBaseExplorer(Loggable, ABC):
         pass
 
     @abstractmethod
-    def _setup_search_highlight(self):
+    def _setup_search_widgets(self):
         """
-        ???+ note "Define how to search and highlight data points."
+        ???+ note "Define how to search data points."
             Left to child classes that have a specific feature format.
         """
         pass
 
+    def _setup_selection_option(self):
+        """
+        ???+ note "Create a group of checkbox(es) for advanced selection options."
+        """
+        from bokeh.models import RadioGroup
+
+        self.selection_option_box = RadioGroup(
+            labels=["keep selecting: none", "union", "intersection", "difference"],
+            active=0,
+        )
+
     def _setup_subset_toggle(self):
         """
         ???+ note "Create a group of buttons for toggling which data subsets to show."
         """
-        from bokeh.models import CheckboxButtonGroup
+        from bokeh.models import CheckboxButtonGroup, Div
+        from bokeh.layouts import column
 
         data_keys = list(self.__class__.SUBSET_GLYPH_KWARGS.keys())
         self.data_key_button_group = CheckboxButtonGroup(
             labels=data_keys, active=list(range(len(data_keys)))
         )
+        self.data_key_button_group_help = Div(text="Toggle data subset display")
+        self.subset_toggle_widget_column = column(
+            self.data_key_button_group_help, self.data_key_button_group
+        )
 
         def update_data_key_display(active):
             visible_keys = {self.data_key_button_group.labels[idx] for idx in active}
@@ -168,7 +203,37 @@ class BokehBaseExplorer(Loggable, ABC):
         )
         self.data_key_button_group.on_click(update_data_key_display)
 
-    def value_patch(self, col_original, col_patch, **kwargs):
+    def _setup_axes_dropdown(self):
+        """
+        ???+ note "Find embedding fields and allow any of them to be set as the x or y axis."
+        """
+        from bokeh.models import Dropdown
+
+        embed_cols = self.find_embedding_fields()
+        init_x, init_y = embed_cols[:2]
+        self.dropdown_x_axis = Dropdown(label=f"X coord: {init_x}", menu=embed_cols)
+        self.dropdown_y_axis = Dropdown(label=f"Y coord: {init_y}", menu=embed_cols)
+
+        def change_x(event):
+            self.dropdown_x_axis.label = f"X coord: {event.item}"
+            for _renderer in self.figure.renderers:
+                _renderer.glyph.x = event.item
+
+        def change_y(event):
+            self.dropdown_y_axis.label = f"Y coord: {event.item}"
+            for _renderer in self.figure.renderers:
+                _renderer.glyph.y = event.item
+
+        self.dropdown_x_axis.on_click(change_x)
+        self.dropdown_y_axis.on_click(change_y)
+
+        # consider allowing dynamic menu refreshment
+        # def refresh_axes_list():
+        #    embed_cols = self.find_embedding_fields()
+        #    self.dropdown_x_axis.menu = embed_cols[:]
+        #    self.dropdown_y_axis.menu = embed_cols[:]
+
+    def value_patch_by_slider(self, col_original, col_patch, **kwargs):
         """
         ???+ note "Allow source values to be dynamically patched through a slider."
             | Param            | Type   | Description                  |
@@ -177,7 +242,7 @@ class BokehBaseExplorer(Loggable, ABC):
             | `col_patch`      | `str`  | column of list of values to use as patches |
             | `**kwargs`       |        | forwarded to the slider |
 
-            [Reference](https://github.com/bokeh/bokeh/blob/2.3.0/examples/howto/patch_app.py)
+            [Reference](https://github.com/bokeh/bokeh/blob/2.4.2/examples/howto/patch_app.py)
         """
         # add a patch slider to widgets, if none exist
         if "patch_slider" not in self._dynamic_widgets:
@@ -230,6 +295,14 @@ class BokehBaseExplorer(Loggable, ABC):
         slider.on_change("value", update_patch)
         self._good(f"Patching {col_original} using {col_patch}")
 
+    def _mandatory_column_defaults(self):
+        """
+        ???+ note "Mandatory columns and default values."
+
+            If default value is None, will raise exception if the column is not found.
+        """
+        return {_col: None for _col in self.__class__.MANDATORY_COLUMNS}
+
     def _setup_dfs(self, df_dict, copy=False):
         """
         ???+ note "Check and store DataFrames **by reference by default**."
@@ -245,6 +318,7 @@ class BokehBaseExplorer(Loggable, ABC):
         expected_keys = set(self.__class__.SUBSET_GLYPH_KWARGS.keys())
 
         # perform high-level df key checks
+        expected_and_supplied = supplied_keys.intersection(expected_keys)
         supplied_not_expected = supplied_keys.difference(expected_keys)
         expected_not_supplied = expected_keys.difference(supplied_keys)
 
@@ -257,27 +331,35 @@ class BokehBaseExplorer(Loggable, ABC):
                 f"{self.__class__.__name__}.__init__(): missing expected df key {_key}"
             )
 
-        # create df with column checks
+        # assign df with column checks
         self.dfs = dict()
-        for _key, _df in df_dict.items():
-            if _key in expected_keys:
-                for _col in self.__class__.MANDATORY_COLUMNS:
-                    if _col not in _df.columns:
-                        # edge case: DataFrame has zero rows
-                        assert (
-                            _df.shape[0] == 0
-                        ), f"Missing column '{_col}' from non-empty {_key} DataFrame: found {list(_df.columns)}"
-                        _df[_col] = None
+        mandatory_col_to_default = self._mandatory_column_defaults()
+        for _key in expected_and_supplied:
+            _df = df_dict[_key]
+            for _col, _default in mandatory_col_to_default.items():
+                # column exists: all good
+                if _col in _df.columns:
+                    continue
+                # no default value: column must be explicitly provided
+                if _default is None:
+                    # edge case: DataFrame has zero rows
+                    _msg = f"Expecting column '{_col}' from {_key} df: found {list(_df.columns)}"
+                    assert _df.shape[0] == 0, _msg
+                # default value available, will use it to create column
+                else:
+                    _df[_col] = _default
+            self.dfs[_key] = _df.copy() if copy else _df
 
-                self.dfs[_key] = _df.copy() if copy else _df
+        # expected dfs must be present
+        for _key in expected_not_supplied:
+            _df = pd.DataFrame(columns=list(mandatory_col_to_default.keys()))
+            self.dfs[_key] = _df
 
     def _setup_sources(self):
         """
         ???+ note "Create, **(not update)**, `ColumnDataSource` objects."
             Intended to be extended in child classes for pre/post processing.
         """
-        from bokeh.events import SelectionGeometry
-
         self._info("Setting up sources")
         self.sources = {_key: ColumnDataSource(_df) for _key, _df in self.dfs.items()}
         self._postprocess_sources()
@@ -285,29 +367,49 @@ class BokehBaseExplorer(Loggable, ABC):
         # initialize attributes that couple with sources
         # extra columns for dynamic plotting
         self._extra_source_cols = defaultdict(dict)
-        # store the last manual selections
-        self._last_selections = {
-            _key: RootUnionFind(set()) for _key in self.sources.keys()
-        }
-        # store commutative, idempotent index filters
-        self._selection_filters = {
-            _key: RootUnionFind(set()) for _key in self.sources.keys()
-        }
 
-        def store_selection(event):
+        self._setup_selection_tools()
+
+    def _setup_subroutine_selection_store(self):
+        """
+        ???+ note "Subroutine of `_setup_selection_tools`."
+            Setup callbacks that interact with manual selections.
+        """
+
+        def store_selection():
             """
             Keep track of the last manual selection.
+            Useful for applying cumulation / filters dynamically.
             """
-            # ensure that nothing happens until the selection event is complete
-            if not event.final:
-                return
+            # determine selection mode
+            selection_option_code = self.selection_option_box.active
 
-            # store selection indices
             for _key, _source in self.sources.items():
                 _selected = _source.selected.indices
-                # use clear() and update() instead of assignment to keep clean references
-                self._last_selections[_key].data.clear()
-                self._last_selections[_key].data.update(_selected)
+                # use sets' in-place methods instead of assignment
+                if selection_option_code == 1:
+                    self._last_selections[_key].data.update(_selected)
+                elif selection_option_code == 2:
+                    self._last_selections[_key].data.intersection_update(_selected)
+                elif selection_option_code == 3:
+                    self._last_selections[_key].data.difference_update(_selected)
+                else:
+                    assert selection_option_code == 0
+                    self._last_selections[_key].data.clear()
+                    self._last_selections[_key].data.update(_selected)
+                _source.selected.indices = list(self._last_selections[_key].data)
+
+        self._store_selection = store_selection
+        self.figure.on_event(
+            SelectionGeometry,
+            lambda event: self._store_selection() if event.final else None,
+        )
+
+    def _setup_subroutine_selection_filter(self):
+        """
+        ???+ note "Subroutine of `_setup_selection_tools`."
+            Setup callbacks that interact with selection filters.
+        """
 
         def trigger_selection_filters(subsets=None):
             """
@@ -326,13 +428,51 @@ class BokehBaseExplorer(Loggable, ABC):
                     _selected = _func(_selected, _key)
                 self.sources[_key].selected.indices = list(_selected)
 
+        # keep reference to trigger_selection_filter() for further access
+        # for example, toggling filters should call the trigger
         self._trigger_selection_filters = trigger_selection_filters
-        self.figure.on_event(SelectionGeometry, store_selection)
         self.figure.on_event(
             SelectionGeometry,
             lambda event: self._trigger_selection_filters() if event.final else None,
         )
 
+    def _setup_subroutine_selection_reset(self):
+        """
+        ???+ note "Subroutine of `_setup_selection_tools`."
+            Setup callbacks for scenarios where the selection should be reset.
+            For example, when the plot sources have changed.
+        """
+
+        def reset_selection():
+            """
+            Clear last manual selections and source selections.
+            Useful during post-processing of refreshed data source.
+            Not directly defined as a method because of `self._last_selections`.
+            """
+            for _key, _source in self.sources.items():
+                self._last_selections[_key].data.clear()
+                _source.selected.indices = []
+
+        self._reset_selection = reset_selection
+
+    def _setup_selection_tools(self):
+        """
+        ???+ note "Create data structures and callbacks for dynamic selections."
+            Useful for linking and filtering selections across explorers.
+        """
+        # store the last manual selections
+        self._last_selections = {
+            _key: RootUnionFind(set()) for _key in self.sources.keys()
+        }
+        # store commutative, idempotent index filters
+        self._selection_filters = {
+            _key: RootUnionFind(set()) for _key in self.sources.keys()
+        }
+
+        self._setup_subroutine_selection_store()
+        self._setup_subroutine_selection_filter()
+        self._setup_subroutine_selection_reset()
+
     def _update_sources(self):
         """
         ???+ note "Update the sources with the corresponding dfs."
@@ -343,7 +483,9 @@ class BokehBaseExplorer(Loggable, ABC):
         for _key in self.dfs.keys():
             self.sources[_key].data = self.dfs[_key]
         self._postprocess_sources()
-        # self._activate_search_builtin(verbose=False)
+
+        # reset selections now that source indices may have changed
+        self._reset_selection()
 
         # reset attribute values that couple with sources
         for _key in self.sources.keys():
@@ -363,42 +505,84 @@ class BokehBaseExplorer(Loggable, ABC):
         """
         pass
 
-    def _activate_search_builtin(self, verbose=True):
+    def activate_search(self):
         """
-        ???+ note "Assign Highlighting callbacks to search results in a manner built into the class."
-            Typically called once during initialization.
+        ???+ note "Assign search response callbacks to search results. Child methods should bind those callbacks to search widgets."
 
-            Note that this is a template method which heavily depends on class attributes.
-            | Param       | Type   | Description                  |
-            | :---------- | :----- | :--------------------------- |
-            | `verbose`   | `bool` | whether to log verbosely     |
+            This is a parent method which takes care of common denominators of parent methods.
+
+            Child methods may inherit the logic here and preprocess/postprocess as needed.
         """
+        self._subroutine_search_create_callbacks()
+        self._subroutine_search_activate_callbacks()
+
+    def _subroutine_search_create_callbacks(self):
+        """
+        ???+ note "Create search callback functions based on feature attributes."
+        """
+        # allow dynamically updated search response through dict element retrieval
+        self._dynamic_callbacks["search_response"] = dict()
+
+        def search_base_response(attr, old, new):
+            for _subset in self.sources.keys():
+                _func = self._dynamic_callbacks["search_response"].get(_subset, blank)
+                _func(attr, old, new)
+            return
+
+        self.search_base_response = search_base_response
+
         for _key, _dict in self.__class__.SUBSET_GLYPH_KWARGS.items():
-            if _key in self.sources.keys():
-                # determine responding attributes
-                _responding = list(_dict["search"].keys())
+            # create a field that holds search results that could be used elsewhere
+            _num_points = len(self.sources[_key].data[self.__class__.PRIMARY_FEATURE])
+            self._extra_source_cols[_key][SEARCH_SCORE_FIELD] = 0
+            self.sources[_key].add([0] * _num_points, SEARCH_SCORE_FIELD)
 
-                # create a field that holds search results that could be used elsewhere
-                _num_points = len(self.sources[_key].data["label"])
-                self._extra_source_cols[_key][SEARCH_SCORE_FIELD] = 0
-                self.sources[_key].add([0] * _num_points, SEARCH_SCORE_FIELD)
+            # make attributes respond to search
+            for _, _params in _dict["search"].items():
+                _updated_kwargs = self._subroutine_search_source_change(
+                    _key,
+                    self.glyph_kwargs[_key],
+                    altered_param=_params,
+                )
+                self.glyph_kwargs[_key].clear()
+                self.glyph_kwargs[_key].update(_updated_kwargs)
 
-                # make attributes respond to search
-                for _flag, _params in _dict["search"].items():
-                    self.glyph_kwargs[_key] = self.activate_search(
-                        _key,
-                        self.glyph_kwargs[_key],
-                        altered_param=_params,
-                    )
-                if verbose:
-                    self._info(
-                        f"Activated {_responding} on subset {_key} to respond to the search widgets."
-                    )
+    def _subroutine_search_activate_callbacks(self):
+        """
+        ???+ note "Activate search callback functions by binding them to widgets."
+        """
+        for _widget in self._search_watch_widgets():
+            _widget.on_change("value", self.search_base_response)
+            self._info(f"activated search base response on {_widget}")
 
     @abstractmethod
-    def activate_search(self, subset, kwargs, altered_param=("size", 10, 5, 7)):
+    def _search_watch_widgets(self):
         """
-        ???+ note "Left to child classes that have a specific feature format."
+        ???+ note "Widgets to trigger search callbacks automatically, which can be different across subclasses."
+
+            Intended for binding callback functions to widgets.
+        """
+
+    @abstractmethod
+    def _validate_search_input(self):
+        """
+        ???+ note "Check the search input, skipping callbacks if it's invalid."
+        """
+        pass
+
+    @abstractmethod
+    def _get_search_score_function(self):
+        """
+        ???+ note "Dynamically create a single-argument scoring function."
+        """
+        pass
+
+    def _subroutine_search_source_change(
+        self, subset, kwargs, altered_param=("size", 10, 5, 7)
+    ):
+        """
+        ???+ note "Subroutine of `activate_search()` on a specific subset."
+            Modifies the plotting source in-place.
 
             | Param           | Type    | Description                   |
             | :-------------- | :------ | :---------------------------  |
@@ -406,7 +590,46 @@ class BokehBaseExplorer(Loggable, ABC):
             | `kwargs`        | `bool`  | kwargs for the plot to add to |
             | `altered_param` | `tuple` | (attribute, positive, negative, default) |
         """
-        pass
+        assert isinstance(kwargs, dict)
+        updated_kwargs = kwargs.copy()
+
+        feature_key = self.__class__.PRIMARY_FEATURE
+        param_key, param_pos, param_neg, param_default = altered_param
+        initial_num = len(self.sources[subset].data[feature_key])
+        self.sources[subset].add([param_default] * initial_num, param_key)
+        self._extra_source_cols[subset][param_key] = param_default
+
+        updated_kwargs[param_key] = param_key
+
+        def score_to_param(score):
+            if score > 0:
+                return param_pos
+            elif score == 0:
+                return param_default
+            else:
+                return param_neg
+
+        def search_response(attr, old, new):
+            valid_flag = self._validate_search_input()
+            if not valid_flag:
+                return
+            score_func = self._get_search_score_function()
+
+            patch_slice = slice(len(self.sources[subset].data[feature_key]))
+            features = self.sources[subset].data[self.__class__.PRIMARY_FEATURE]
+            # score_func is potentially slow; track its progress
+            # search_scores = list(map(score_func, tqdm(features, desc="Search score")))
+            search_scores = list(map(score_func, features))
+            search_params = list(map(score_to_param, search_scores))
+            self.sources[subset].patch(
+                {SEARCH_SCORE_FIELD: [(patch_slice, search_scores)]}
+            )
+            self.sources[subset].patch({param_key: [(patch_slice, search_params)]})
+            return
+
+        # assign dynamic callback
+        self._dynamic_callbacks["search_response"][subset] = search_response
+        return updated_kwargs
 
     def _prelink_check(self, other):
         """
@@ -431,9 +654,6 @@ class BokehBaseExplorer(Loggable, ABC):
         # link selection in a bidirectional manner
         sl, sr = self.sources[key], other.sources[other_key]
 
-        # deprecated: use js_link to sync attributes
-        # sl.selected.js_link("indices", sr.selected, "indices")
-        # sr.selected.js_link("indices", sl.selected, "indices")
         def left_to_right(attr, old, new):
             sr.selected.indices = sl.selected.indices[:]
 
@@ -452,6 +672,23 @@ class BokehBaseExplorer(Loggable, ABC):
         )
         self._selection_filters[key].union(other._selection_filters[other_key])
 
+    def link_selection_options(self, other):
+        """
+        ???+ note "Synchronize the selection option values between explorers."
+            | Param   | Type    | Description                    |
+            | :------ | :------ | :----------------------------- |
+            | `other` | `BokehBaseExplorer` | the other explorer |
+        """
+
+        def left_to_right(attr, old, new):
+            other.selection_option_box.active = self.selection_option_box.active
+
+        def right_to_left(attr, old, new):
+            self.selection_option_box.active = other.selection_option_box.active
+
+        self.selection_option_box.on_change("active", left_to_right)
+        other.selection_option_box.on_change("active", right_to_left)
+
     def link_xy_range(self, other):
         """
         ???+ note "Synchronize plotting ranges on the xy-plane."
@@ -479,6 +716,29 @@ class BokehBaseExplorer(Loggable, ABC):
         """
         pass
 
+    def find_embedding_fields(self):
+        """
+        ???+ note "Find embedding fields from dataframes."
+
+            Intended for scenarios where the embedding is higher than two-dimensional.
+        """
+        embedding_cols = None
+        for _key, _df in self.dfs.items():
+            # edge case: dataframe is empty
+            if _df.shape[0] == 0:
+                continue
+            # automatically find embedding columns
+            _emb_cols = sorted(filter(is_embedding_field, _df.columns))
+            if embedding_cols is None:
+                embedding_cols = _emb_cols
+            else:
+                # embedding columns must be the same across subsets
+                assert embedding_cols == _emb_cols, "Inconsistent embedding columns"
+        assert (
+            len(embedding_cols) >= 2
+        ), f"Expected at least two embedding columns, found {embedding_cols}"
+        return embedding_cols
+
     def auto_color_mapping(self):
         """
         ???+ note "Find all labels and an appropriate color for each."
