@@ -2,7 +2,8 @@
 ???+ note "Intermediate classes based on the functionality."
 """
 import numpy as np
-from bokeh.models import CDSView, IndexFilter
+from collections import OrderedDict
+from bokeh.models import CDSView, IndexFilter, Dropdown, Button
 from bokeh.palettes import Category20
 from bokeh.layouts import row
 from hover import module_config
@@ -18,7 +19,8 @@ class BokehDataFinder(BokehBaseExplorer):
 
         Features:
 
-        -   the search widgets will highlight the results through a change of color, which gives arguably the best focus.
+        -   the search widgets will highlight the results through a change of color.
+        -   the search results can be used as a filter condition.
     """
 
     SUBSET_GLYPH_KWARGS = {
@@ -45,37 +47,37 @@ class BokehDataFinder(BokehBaseExplorer):
             labels=["use as selection filter"], active=[]
         )
 
-        def activated():
+    def _subroutine_search_activate_callbacks(self):
+        """
+        ???+ note "Activate search callback functions by binding them to widgets."
+        """
+        super()._subroutine_search_activate_callbacks()
+
+        def filter_flag():
             return bool(0 in self.search_filter_box.active)
 
         def filter_by_search(indices, subset):
             """
             Filter selection with search results on a subset.
             """
+            if not filter_flag():
+                return indices
+
             search_scores = self.sources[subset].data[SEARCH_SCORE_FIELD]
             matched = set(np.where(np.array(search_scores) > 0)[0])
             return indices.intersection(matched)
 
         for _key in self.sources.keys():
-            self._selection_filters[_key].data.add(
-                lambda indices, subset: filter_by_search(indices, subset)
-                if activated()
-                else indices
-            )
+            self._selection_filters[_key].data.add(filter_by_search)
 
         # when toggled as active, search changes trigger selection filter
-        self.search_pos.on_change(
-            "value",
-            lambda attr, old, new: self._trigger_selection_filters()
-            if activated()
-            else None,
-        )
-        self.search_neg.on_change(
-            "value",
-            lambda attr, old, new: self._trigger_selection_filters()
-            if activated()
-            else None,
-        )
+        for _widget in self._search_watch_widgets():
+            _widget.on_change(
+                "value",
+                lambda attr, old, new: self._trigger_selection_filters()
+                if filter_flag()
+                else None,
+            )
 
         # active toggles always trigger selection filter
         self.search_filter_box.on_change(
@@ -86,9 +88,10 @@ class BokehDataFinder(BokehBaseExplorer):
         """
         ???+ note "Plot all data points."
         """
+        xy_axes = self.find_embedding_fields()[:2]
         for _key, _source in self.sources.items():
             self.figure.circle(
-                "x", "y", name=_key, source=_source, **self.glyph_kwargs[_key]
+                *xy_axes, name=_key, source=_source, **self.glyph_kwargs[_key]
             )
             self._good(f"Plotted subset {_key} with {self.dfs[_key].shape[0]} points")
 
@@ -116,6 +119,8 @@ class BokehDataAnnotator(BokehBaseExplorer):
     def _postprocess_sources(self):
         """
         ???+ note "Infer glyph colors from the label dynamically."
+
+            This is during initialization or re-plotting, creating a new attribute column for each data source.
         """
         color_dict = self.auto_color_mapping()
 
@@ -127,11 +132,30 @@ class BokehDataAnnotator(BokehBaseExplorer):
             )
             self.sources[_key].add(_color, SOURCE_COLOR_FIELD)
 
+    def _update_colors(self):
+        """
+        ???+ note "Infer glyph colors from the label dynamically."
+
+            This is during annotation callbacks, patching an existing column for the `raw` subset only.
+        """
+        # infer glyph colors dynamically
+        color_dict = self.auto_color_mapping()
+
+        color_list = (
+            self.dfs["raw"]["label"]
+            .apply(lambda label: color_dict.get(label, "gainsboro"))
+            .tolist()
+        )
+        self.sources["raw"].patch(
+            {SOURCE_COLOR_FIELD: [(slice(len(color_list)), color_list)]}
+        )
+        self._good(f"Updated annotator plot at {current_time()}")
+
     def _setup_widgets(self):
         """
         ???+ note "Create annotator widgets and assign Python callbacks."
         """
-        from bokeh.models import TextInput, Button
+        from bokeh.models import TextInput
 
         super()._setup_widgets()
 
@@ -153,30 +177,18 @@ class BokehDataAnnotator(BokehBaseExplorer):
             selected_idx = self.sources["raw"].selected.indices
             if not selected_idx:
                 self._warn(
-                    "Attempting annotation: did not select any data points. Eligible subset is 'raw'."
+                    "attempting annotation: did not select any data points. Eligible subset is 'raw'."
                 )
                 return
+            self._info(f"applying {len(selected_idx)} annotations...")
 
-            self._info(f"Applying {len(selected_idx)} annotations: {label}")
             # update label in both the df and the data source
-            self.dfs["raw"].at[selected_idx, "label"] = label
-            for _idx in selected_idx:
-                _idx = int(_idx)
-                self.sources["raw"].patch({"label": [(_idx, label)]})
-            self._good(f"Applied {len(selected_idx)} annotations: {label}")
+            self.dfs["raw"].loc[selected_idx, "label"] = label
+            patch_to_apply = [(_idx, label) for _idx in selected_idx]
+            self.sources["raw"].patch({"label": patch_to_apply})
+            self._good(f"applied {len(selected_idx)} annotations: {label}")
 
-            # infer glyph colors dynamically
-            color_dict = self.auto_color_mapping()
-
-            color_list = (
-                self.dfs["raw"]["label"]
-                .apply(lambda label: color_dict.get(label, "gainsboro"))
-                .tolist()
-            )
-            self.sources["raw"].patch(
-                {SOURCE_COLOR_FIELD: [(slice(len(color_list)), color_list)]}
-            )
-            self._good(f"Updated annotator plot at {current_time()}")
+            self._update_colors()
 
         # assign the callback and keep the reference
         self._callback_apply = callback_apply
@@ -189,10 +201,10 @@ class BokehDataAnnotator(BokehBaseExplorer):
             Overrides the parent method.
             Determines the label -> color mapping dynamically.
         """
+        xy_axes = self.find_embedding_fields()[:2]
         for _key, _source in self.sources.items():
             self.figure.circle(
-                "x",
-                "y",
+                *xy_axes,
                 name=_key,
                 color=SOURCE_COLOR_FIELD,
                 source=_source,
@@ -216,6 +228,7 @@ class BokehSoftLabelExplorer(BokehBaseExplorer):
         _key: {"constant": {"line_alpha": 0.5}, "search": {"size": ("size", 10, 5, 7)}}
         for _key in ["raw", "train", "dev"]
     }
+    DEFAULT_SUBSET_MAPPING = {_k: _k for _k in ["raw", "train", "dev"]}
 
     def __init__(self, df_dict, label_col, score_col, **kwargs):
         """
@@ -250,21 +263,20 @@ class BokehSoftLabelExplorer(BokehBaseExplorer):
         )
         return f"{standard}\n{extra}"
 
-    def _setup_dfs(self, df_dict, **kwargs):
+    def _mandatory_column_defaults(self):
         """
-        ???+ note "On top of the parent method, add filler values to additional columns."
-            | Param      | Type   | Description                  |
-            | :--------- | :----- | :--------------------------- |
-            | `df_dict`  | `dict` | `str` -> `DataFrame` mapping |
-            | `**kwargs` |        | forwarded to the parent method |
-        """
-        super()._setup_dfs(df_dict, **kwargs)
+        ???+ note "Mandatory columns and default values."
 
-        for _key, _df in self.dfs.items():
-            if self.label_col not in _df.columns:
-                _df[self.label_col] = module_config.ABSTAIN_DECODED
-            if self.score_col not in _df.columns:
-                _df[self.score_col] = 0.5
+            If default value is None, will raise exception if the column is not found.
+        """
+        column_to_value = super()._mandatory_column_defaults()
+        column_to_value.update(
+            {
+                self.label_col: module_config.ABSTAIN_DECODED,
+                self.score_col: 0.5,
+            }
+        )
+        return column_to_value
 
     def _postprocess_sources(self):
         """
@@ -318,7 +330,7 @@ class BokehSoftLabelExplorer(BokehBaseExplorer):
         )
         self.score_filter = row(self.score_range, self.score_filter_box)
 
-        def activated():
+        def filter_flag():
             return bool(0 in self.score_filter_box.active)
 
         def subroutine(df, lower, upper):
@@ -334,22 +346,21 @@ class BokehSoftLabelExplorer(BokehBaseExplorer):
             """
             Filter selection with slider range on a subset.
             """
+            if not filter_flag():
+                return indices
+
             in_range = subroutine(self.dfs[subset], *self.score_range.value)
             return indices.intersection(in_range)
 
         # selection change triggers score filter on the changed subset IFF filter box is toggled
         for _key in self.sources.keys():
-            self._selection_filters[_key].data.add(
-                lambda indices, subset: filter_by_score(indices, subset)
-                if activated()
-                else indices
-            )
+            self._selection_filters[_key].data.add(filter_by_score)
 
         # when toggled as active, score range change triggers selection filter
         self.score_range.on_change(
             "value",
             lambda attr, old, new: self._trigger_selection_filters()
-            if activated()
+            if filter_flag()
             else None,
         )
 
@@ -365,6 +376,7 @@ class BokehSoftLabelExplorer(BokehBaseExplorer):
             | :--------- | :----- | :--------------------------- |
             | `**kwargs` |        | forwarded to plotting markers |
         """
+        xy_axes = self.find_embedding_fields()[:2]
         for _key, _source in self.sources.items():
             # prepare plot settings
             preset_kwargs = {
@@ -375,7 +387,7 @@ class BokehSoftLabelExplorer(BokehBaseExplorer):
             eff_kwargs.update(preset_kwargs)
             eff_kwargs.update(kwargs)
 
-            self.figure.circle("x", "y", name=_key, source=_source, **eff_kwargs)
+            self.figure.circle(*xy_axes, name=_key, source=_source, **eff_kwargs)
             self._good(f"Plotted subset {_key} with {self.dfs[_key].shape[0]} points")
 
 
@@ -397,6 +409,7 @@ class BokehMarginExplorer(BokehBaseExplorer):
         }
         for _key in ["raw", "train", "dev"]
     }
+    DEFAULT_SUBSET_MAPPING = {_k: _k for _k in ["raw", "train", "dev"]}
 
     def __init__(self, df_dict, label_col_a, label_col_b, **kwargs):
         """
@@ -417,21 +430,20 @@ class BokehMarginExplorer(BokehBaseExplorer):
         self.label_col_b = label_col_b
         super().__init__(df_dict, **kwargs)
 
-    def _setup_dfs(self, df_dict, **kwargs):
+    def _mandatory_column_defaults(self):
         """
-        ???+ note "On top of the parent method, add column checks."
-            | Param      | Type   | Description                  |
-            | :--------- | :----- | :--------------------------- |
-            | `df_dict`  | `dict` | `str` -> `DataFrame` mapping |
-            | `**kwargs` |        | forwarded to the parent method |
-        """
-        super()._setup_dfs(df_dict, **kwargs)
+        ???+ note "Mandatory columns and default values."
 
-        for _key, _df in self.dfs.items():
-            for _col in [self.label_col_a, self.label_col_b]:
-                assert (
-                    _col in _df.columns
-                ), f"Expected column {_col} among {list(_df.columns)}"
+            If default value is None, will raise exception if the column is not found.
+        """
+        column_to_value = super()._mandatory_column_defaults()
+        column_to_value.update(
+            {
+                self.label_col_a: None,
+                self.label_col_b: None,
+            }
+        )
+        return column_to_value
 
     def plot(self, label, **kwargs):
         """
@@ -442,6 +454,7 @@ class BokehMarginExplorer(BokehBaseExplorer):
             | `**kwargs` |        | forwarded to plotting markers |
         """
 
+        xy_axes = self.find_embedding_fields()[:2]
         for _key, _source in self.sources.items():
             # prepare plot settings
             eff_kwargs = self.glyph_kwargs[_key].copy()
@@ -473,7 +486,7 @@ class BokehMarginExplorer(BokehBaseExplorer):
             for _dict in to_plot:
                 _view = _dict["view"]
                 _marker = _dict["marker"]
-                _marker("x", "y", name=_key, source=_source, view=_view, **eff_kwargs)
+                _marker(*xy_axes, name=_key, source=_source, view=_view, **eff_kwargs)
 
 
 class BokehSnorkelExplorer(BokehBaseExplorer):
@@ -503,6 +516,7 @@ class BokehSnorkelExplorer(BokehBaseExplorer):
             "search": {"size": ("size", 10, 5, 7)},
         },
     }
+    DEFAULT_SUBSET_MAPPING = {"raw": "raw", "dev": "labeled"}
 
     def __init__(self, df_dict, **kwargs):
         """
@@ -518,25 +532,279 @@ class BokehSnorkelExplorer(BokehBaseExplorer):
             | `**kwargs`  |        | forwarded to `bokeh.plotting.figure` |
         """
         super().__init__(df_dict, **kwargs)
+        self.palette = list(Category20[20])
+        self._subscribed_lf_list = None
 
-        # initialize a list to keep track of plotted LFs
-        self.lfs = []
-        self.palette = Category20[20]
+    def _setup_sources(self):
+        """
+        ???+ note "Create data structures that source interactions will need."
+        """
+        # keep track of plotted LFs and glyphs, which will interact with sources
+        self.lf_data = OrderedDict()
+        super()._setup_sources()
+
+    @property
+    def subscribed_lf_list(self):
+        """
+        ???+ note "A list of LFs to which the explorer can be lazily synchronized."
+
+            Intended for recipes where the user can modify LFs without having access to the explorer.
+        """
+        return self._subscribed_lf_list
+
+    @subscribed_lf_list.setter
+    def subscribed_lf_list(self, lf_list):
+        """
+        ???+ note "Subscribe to a list of LFs."
+        """
+        assert isinstance(lf_list, list), f"Expected a list of LFs, got {lf_list}"
+        if self.subscribed_lf_list is None:
+            self._good("Subscribed to a labeling function list BY REFERENCE.")
+        else:
+            self._warn("Changing labeling function list subscription.")
+        self._subscribed_lf_list = lf_list
+        self._callback_refresh_lf_menu()
+
+    def _setup_widgets(self):
+        """
+        ???+ note "Create labeling function support widgets and assign Python callbacks."
+        """
+        super()._setup_widgets()
+        self._subroutine_setup_lf_list_refresher()
+        self._subroutine_setup_lf_apply_trigger()
+        self._subroutine_setup_lf_filter_trigger()
+
+    def _subroutine_setup_lf_list_refresher(self):
+        """
+        ???+ note "Create widget for refreshing LF list and replotting."
+        """
+        self.lf_list_refresher = Button(
+            label="Refresh Functions",
+            height_policy="fit",
+            width_policy="min",
+        )
+
+        def callback_refresh_lf_plot():
+            """
+            Re-plot according to subscribed_lf_list.
+            """
+            if self.subscribed_lf_list is None:
+                self._warn("cannot refresh LF plot without subscribed LF list.")
+                return
+            lf_names_to_keep = set([_lf.name for _lf in self.subscribed_lf_list])
+            lf_names_to_drop = set(self.lf_data.keys()).difference(lf_names_to_keep)
+            for _lf_name in lf_names_to_drop:
+                self.unplot_lf(_lf_name)
+            for _lf in self.subscribed_lf_list:
+                self.plot_lf(_lf)
+
+        def callback_refresh_lf_menu():
+            """
+            The menu was assigned by value and needs to stay consistent with LF updates.
+            To be triggered in self.plot_new_lf() and self.unplot_lf().
+            """
+            self.lf_apply_trigger.menu = list(self.lf_data.keys())
+            self.lf_filter_trigger.menu = list(self.lf_data.keys())
+
+        self._callback_refresh_lf_menu = callback_refresh_lf_menu
+        self.lf_list_refresher.on_click(callback_refresh_lf_plot)
+        # self.lf_list_refresher.on_click(callback_refresh_lf_menu)
+
+    def _subroutine_setup_lf_apply_trigger(self):
+        """
+        ???+ note "Create widget for applying LFs on data."
+        """
+        self.lf_apply_trigger = Dropdown(
+            label="Apply Labels",
+            button_type="warning",
+            menu=list(self.lf_data.keys()),
+            height_policy="fit",
+            width_policy="min",
+        )
+
+        def callback_apply(event):
+            """
+            A callback on clicking the 'self.lf_apply_trigger' button.
+
+            Update labels in the source similarly to the annotator.
+            However, in this explorer, because LFs already use color, the produced labels will not.
+            """
+            lf = self.lf_data[event.item]["lf"]
+            assert callable(lf), f"Expected a function, got {lf}"
+
+            selected_idx = self.sources["raw"].selected.indices
+            if not selected_idx:
+                self._warn(
+                    "attempting labeling by function: did not select any data points. Eligible subset is 'raw'."
+                )
+                return
+
+            labels = self.dfs["raw"].iloc[selected_idx].apply(lf, axis=1).values
+            num_nontrivial = len(
+                list(filter(lambda l: l != module_config.ABSTAIN_DECODED, labels))
+            )
+
+            # update label in both the df and the data source
+            self.dfs["raw"].loc[selected_idx, "label"] = labels
+            for _idx, _label in zip(selected_idx, labels):
+                _idx = int(_idx)
+                self.sources["raw"].patch({"label": [(_idx, _label)]})
+            self._info(
+                f"applied {num_nontrivial}/{len(labels)} annotations by func {lf.name}"
+            )
+
+        self.lf_apply_trigger.on_click(callback_apply)
+
+    def _subroutine_setup_lf_filter_trigger(self):
+        """
+        ???+ note "Create widget for using LFs to filter data."
+        """
+        self.lf_filter_trigger = Dropdown(
+            label="Use as Selection Filter",
+            button_type="primary",
+            menu=list(self.lf_data.keys()),
+            height_policy="fit",
+            width_policy="min",
+        )
+
+        def callback_filter(event):
+            """
+            A callback on clicking the 'self.lf_filter_trigger' button.
+
+            Update selected indices in a one-time manner.
+            """
+            lf = self.lf_data[event.item]["lf"]
+            assert callable(lf), f"Expected a function, got {lf}"
+
+            for _key, _source in self.sources.items():
+                _selected = _source.selected.indices
+                _labels = self.dfs[_key].iloc[_selected].apply(lf, axis=1).values
+                _kept = [
+                    _idx
+                    for _idx, _label in zip(_selected, _labels)
+                    if _label != module_config.ABSTAIN_DECODED
+                ]
+                self.sources[_key].selected.indices = _kept
+
+        self.lf_filter_trigger.on_click(callback_filter)
+
+    def _postprocess_sources(self):
+        """
+        ???+ note "Refresh all LF glyphs because data source has changed."
+        """
+        for _lf_name in self.lf_data.keys():
+            self.refresh_glyphs(_lf_name)
 
     def plot(self, *args, **kwargs):
         """
         ???+ note "Plot the raw subset in the background."
         """
+        xy_axes = self.find_embedding_fields()[:2]
         self.figure.circle(
-            "x", "y", name="raw", source=self.sources["raw"], **self.glyph_kwargs["raw"]
+            *xy_axes, name="raw", source=self.sources["raw"], **self.glyph_kwargs["raw"]
         )
         self._good(f"Plotted subset raw with {self.dfs['raw'].shape[0]} points")
 
-    def plot_lf(
+    def plot_lf(self, lf, **kwargs):
+        """
+        ???+ note "Add or refresh a single labeling function on the plot."
+            | Param       | Type             | Description                  |
+            | :---------- | :--------------- | :--------------------------- |
+            | `lf`        | `callable`       | labeling function decorated by `@labeling_function()` from `hover.utils.snorkel_helper` |
+            | `**kwargs`  |             | forwarded to `self.plot_new_lf()` |
+        """
+        # keep track of added LF
+        if lf.name in self.lf_data:
+            # skip if the functions are identical
+            if self.lf_data[lf.name]["lf"] is lf:
+                return
+            # overwrite the function and refresh glyphs
+            self.lf_data[lf.name]["lf"] = lf
+            self.refresh_glyphs(lf.name)
+            return
+
+        self.plot_new_lf(lf, **kwargs)
+
+    def unplot_lf(self, lf_name):
+        """
+        ???+ note "Remove a single labeling function from the plot."
+            | Param     | Type   | Description               |
+            | :-------- | :----- | :------------------------ |
+            | `lf_name` | `str`  | name of labeling function |
+        """
+        assert lf_name in self.lf_data, f"trying to remove non-existing LF: {lf_name}"
+
+        data_dict = self.lf_data.pop(lf_name)
+        lf, glyph_dict = data_dict["lf"], data_dict["glyphs"]
+        assert lf.name == lf_name, f"LF name mismatch: {lf.name} vs {lf_name}"
+
+        # remove from legend, checking that there is exactly one entry
+        legend_idx_to_pop = None
+        for i, _item in enumerate(self.figure.legend.items):
+            _label = _item.label.get("value", "")
+            if _label == lf_name:
+                assert legend_idx_to_pop is None, f"Legend collision: {lf_name}"
+                legend_idx_to_pop = i
+        assert isinstance(legend_idx_to_pop, int), f"Missing from legend: {lf_name}"
+        self.figure.legend.items.pop(legend_idx_to_pop)
+
+        # remove from renderers
+        # get indices to pop in ascending order
+        renderer_indices_to_pop = []
+        for i, _renderer in enumerate(self.figure.renderers):
+            if lf_name in _renderer.glyph.tags:
+                renderer_indices_to_pop.append(i)
+        # check that the number of glyphs founded matches expected value
+        num_fnd, num_exp = len(renderer_indices_to_pop), len(glyph_dict)
+        assert num_fnd == num_exp, f"Glyph mismatch: {num_fnd} vs. {num_exp}"
+        # process indices in descending order to avoid shifts
+        for i in renderer_indices_to_pop[::-1]:
+            self.figure.renderers.pop(i)
+
+        # return color to palette so that another LF can use it
+        self.palette.append(data_dict["color"])
+
+        self._callback_refresh_lf_menu()
+        self._good(f"Unplotted LF {lf_name}")
+
+    def refresh_glyphs(self, lf_name):
+        """
+        ???+ note "Refresh the glyph(s) of a single LF based on its name."
+            | Param     | Type   | Description               |
+            | :-------- | :----- | :------------------------ |
+            | `lf_name` | `str`  | name of labeling function |
+
+            Assumes that specified C/I/M/H glyphs are stored.
+            1. re-compute L_raw/L_labeled and CDSViews
+            2. update the view for each glyph
+        """
+        assert lf_name in self.lf_data, f"trying to refresh non-existing LF: {lf_name}"
+
+        lf = self.lf_data[lf_name]["lf"]
+        L_raw = self.dfs["raw"].apply(lf, axis=1).values
+        L_labeled = self.dfs["labeled"].apply(lf, axis=1).values
+
+        glyph_codes = self.lf_data[lf_name]["glyphs"].keys()
+        if "C" in glyph_codes:
+            c_view = self._view_correct(L_labeled)
+            self.lf_data[lf_name]["glyphs"]["C"].view = c_view
+        if "I" in glyph_codes:
+            i_view = self._view_incorrect(L_labeled)
+            self.lf_data[lf_name]["glyphs"]["I"].view = i_view
+        if "M" in glyph_codes:
+            m_view = self._view_missed(L_labeled, lf.targets)
+            self.lf_data[lf_name]["glyphs"]["M"].view = m_view
+        if "H" in glyph_codes:
+            h_view = self._view_hit(L_raw)
+            self.lf_data[lf_name]["glyphs"]["H"].view = h_view
+
+        self._good(f"Refreshed the glyphs of LF {lf_name}")
+
+    def plot_new_lf(
         self, lf, L_raw=None, L_labeled=None, include=("C", "I", "M"), **kwargs
     ):
         """
-        ???+ note "Plot about a single labeling function."
+        ???+ note "Plot a single labeling function and keep its settings for update."
             | Param       | Type             | Description                  |
             | :---------- | :--------------- | :--------------------------- |
             | `lf`        | `callable`       | labeling function decorated by `@labeling_function()` from `hover.utils.snorkel_helper` |
@@ -547,12 +815,12 @@ class BokehSnorkelExplorer(BokehBaseExplorer):
 
 
             - lf: labeling function decorated by `@labeling_function()` from `hover.utils.snorkel_helper`
-            - L_raw: .
-            - L_labeled: .
+            - L_raw: numpy.ndarray
+            - L_labeled: numpy.ndarray
             - include: subsets to show, which can be correct(C)/incorrect(I)/missed(M)/hit(H).
         """
-        # keep track of added LF
-        self.lfs.append(lf)
+        # existing LF should not trigger this method
+        assert lf.name not in self.lf_data, f"LF collision: {lf.name}"
 
         # calculate predicted labels if not provided
         if L_raw is None:
@@ -561,8 +829,10 @@ class BokehSnorkelExplorer(BokehBaseExplorer):
             L_labeled = self.dfs["labeled"].apply(lf, axis=1).values
 
         # prepare plot settings
-        legend_label = f"{', '.join(lf.targets)} | {lf.name}"
-        color = self.palette[len(self.lfs) - 1]
+        assert self.palette, f"Palette depleted, # LFs: {len(self.lf_data)}"
+        legend_label = lf.name
+        color = self.palette.pop(0)
+        xy_axes = self.find_embedding_fields()[:2]
 
         raw_glyph_kwargs = self.glyph_kwargs["raw"].copy()
         raw_glyph_kwargs["legend_label"] = legend_label
@@ -574,52 +844,57 @@ class BokehSnorkelExplorer(BokehBaseExplorer):
         labeled_glyph_kwargs["color"] = color
         labeled_glyph_kwargs.update(kwargs)
 
-        # create correct/incorrect/missed/hit subsets
-        to_plot = []
+        # create dictionary to prepare for dynamic lf & glyph updates
+        data_dict = {"lf": lf, "color": color, "glyphs": {}}
+
+        # add correct/incorrect/missed/hit glyphs
         if "C" in include:
-            to_plot.append(
-                {
-                    "name": "labeled",
-                    "view": self._view_correct(L_labeled),
-                    "marker": self.figure.square,
-                    "kwargs": labeled_glyph_kwargs,
-                }
+            view = self._view_correct(L_labeled)
+            data_dict["glyphs"]["C"] = self.figure.square(
+                *xy_axes,
+                source=view.source,
+                view=view,
+                name="labeled",
+                tags=[lf.name],
+                **labeled_glyph_kwargs,
             )
         if "I" in include:
-            to_plot.append(
-                {
-                    "name": "labeled",
-                    "view": self._view_incorrect(L_labeled),
-                    "marker": self.figure.x,
-                    "kwargs": labeled_glyph_kwargs,
-                }
+            view = self._view_incorrect(L_labeled)
+            data_dict["glyphs"]["I"] = self.figure.x(
+                *xy_axes,
+                source=view.source,
+                view=view,
+                name="labeled",
+                tags=[lf.name],
+                **labeled_glyph_kwargs,
             )
         if "M" in include:
-            to_plot.append(
-                {
-                    "name": "labeled",
-                    "view": self._view_missed(L_labeled, lf.targets),
-                    "marker": self.figure.cross,
-                    "kwargs": labeled_glyph_kwargs,
-                }
+            view = self._view_missed(L_labeled, lf.targets)
+            data_dict["glyphs"]["M"] = self.figure.cross(
+                *xy_axes,
+                source=view.source,
+                view=view,
+                name="labeled",
+                tags=[lf.name],
+                **labeled_glyph_kwargs,
             )
         if "H" in include:
-            to_plot.append(
-                {
-                    "name": "raw",
-                    "view": self._view_hit(L_raw),
-                    "marker": self.figure.circle,
-                    "kwargs": raw_glyph_kwargs,
-                }
+            view = self._view_hit(L_raw)
+            data_dict["glyphs"]["H"] = self.figure.circle(
+                *xy_axes,
+                source=view.source,
+                view=view,
+                name="raw",
+                tags=[lf.name],
+                **raw_glyph_kwargs,
             )
 
-        # plot created subsets
-        for _dict in to_plot:
-            _name = _dict["name"]
-            _view = _dict["view"]
-            _marker = _dict["marker"]
-            _kwargs = _dict["kwargs"]
-            _marker("x", "y", source=_view.source, view=_view, name=_name, **_kwargs)
+        # assign the completed dictionary
+        self.lf_data[lf.name] = data_dict
+        # reflect LF update in widgets
+        self._callback_refresh_lf_menu()
+
+        self._good(f"Plotted new LF {lf.name}")
 
     def _view_correct(self, L_labeled):
         """
@@ -628,9 +903,12 @@ class BokehSnorkelExplorer(BokehBaseExplorer):
             | :---------- | :----------- | :--------------------------- |
             | `L_labeled` | `np.ndarray` | predictions on the labeled subset |
         """
-        agreed = self.dfs["labeled"]["label"].values == L_labeled
-        attempted = L_labeled != module_config.ABSTAIN_DECODED
-        indices = np.where(np.multiply(agreed, attempted))[0].tolist()
+        if L_labeled.shape[0] == 0:
+            indices = []
+        else:
+            agreed = self.dfs["labeled"]["label"].values == L_labeled
+            attempted = L_labeled != module_config.ABSTAIN_DECODED
+            indices = np.where(np.multiply(agreed, attempted))[0].tolist()
         view = CDSView(source=self.sources["labeled"], filters=[IndexFilter(indices)])
         return view
 
@@ -641,9 +919,12 @@ class BokehSnorkelExplorer(BokehBaseExplorer):
             | :---------- | :----------- | :--------------------------- |
             | `L_labeled` | `np.ndarray` | predictions on the labeled subset |
         """
-        disagreed = self.dfs["labeled"]["label"].values != L_labeled
-        attempted = L_labeled != module_config.ABSTAIN_DECODED
-        indices = np.where(np.multiply(disagreed, attempted))[0].tolist()
+        if L_labeled.shape[0] == 0:
+            indices = []
+        else:
+            disagreed = self.dfs["labeled"]["label"].values != L_labeled
+            attempted = L_labeled != module_config.ABSTAIN_DECODED
+            indices = np.where(np.multiply(disagreed, attempted))[0].tolist()
         view = CDSView(source=self.sources["labeled"], filters=[IndexFilter(indices)])
         return view
 
@@ -655,9 +936,12 @@ class BokehSnorkelExplorer(BokehBaseExplorer):
             | `L_labeled` | `np.ndarray`  | predictions on the labeled subset |
             | `targets` | `list` of `str` | labels that the function aims for |
         """
-        targetable = np.isin(self.dfs["labeled"]["label"], targets)
-        abstained = L_labeled == module_config.ABSTAIN_DECODED
-        indices = np.where(np.multiply(targetable, abstained))[0].tolist()
+        if L_labeled.shape[0] == 0:
+            indices = []
+        else:
+            targetable = np.isin(self.dfs["labeled"]["label"], targets)
+            abstained = L_labeled == module_config.ABSTAIN_DECODED
+            indices = np.where(np.multiply(targetable, abstained))[0].tolist()
         view = CDSView(source=self.sources["labeled"], filters=[IndexFilter(indices)])
         return view
 
@@ -668,6 +952,9 @@ class BokehSnorkelExplorer(BokehBaseExplorer):
             | :---------- | :----------- | :--------------------------- |
             | `L_raw`     | `np.ndarray` | predictions on the raw subset |
         """
-        indices = np.where(L_raw != module_config.ABSTAIN_DECODED)[0].tolist()
+        if L_raw.shape[0] == 0:
+            indices = []
+        else:
+            indices = np.where(L_raw != module_config.ABSTAIN_DECODED)[0].tolist()
         view = CDSView(source=self.sources["raw"], filters=[IndexFilter(indices)])
         return view
