@@ -1,9 +1,10 @@
 """
 ???+ note "Useful subroutines for working with bokeh in general."
 """
-import warnings
+import os
 from functools import wraps
 from traceback import format_exc
+from urllib.parse import urljoin, urlparse
 from bokeh.models import PreText
 from bokeh.layouts import column
 from bokeh.palettes import Category10, Category20
@@ -92,18 +93,15 @@ def servable(title=None):
                 layout = column(spinner)
 
                 def progress():
-                    spinner.text += "."
+                    """
+                    If still loading, show some progress.
+                    """
+                    if spinner in layout.children:
+                        spinner.text += "."
 
                 def load():
                     try:
                         bokeh_model = func(*args, **kwargs)
-                        # remove spinner and its update
-                        try:
-                            doc.remove_periodic_callback(progress)
-                        except Exception as e:
-                            warnings.warn(
-                                f"@servable: trying to remove periodic callback, got {type(e)}: {e}"
-                            )
                         layout.children.append(bokeh_model)
                         layout.children.pop(0)
                     except Exception as e:
@@ -123,6 +121,23 @@ def servable(title=None):
     return wrapper
 
 
+def show_as_interactive(obj, **kwargs):
+    """
+    ???+ note "Wrap a bokeh LayoutDOM as an application to allow Python callbacks."
+
+        Must have the same signature as `bokeh.io.show()`[https://docs.bokeh.org/en/latest/docs/reference/io.html#bokeh.io.show].
+    """
+    from bokeh.io import show
+    from bokeh.models.layouts import LayoutDOM
+
+    assert isinstance(obj, LayoutDOM), f"Expected Bokeh LayoutDOM, got {type(obj)}"
+
+    def handle(doc):
+        doc.add_root(column(obj))
+
+    return show(handle, **kwargs)
+
+
 def bokeh_hover_tooltip(
     label=False,
     text=False,
@@ -135,13 +150,13 @@ def bokeh_hover_tooltip(
     """
     ???+ note "Create a Bokeh hover tooltip from a template."
 
-        - param label: whether to expect and show a "label" field.
-        - param text: whether to expect and show a "text" field.
-        - param image: whether to expect and show an "image" (url/path) field.
-        - param audio: whether to expect and show an "audio" (url/path) field.
-        - param coords: whether to show xy-coordinates.
-        - param index: whether to show indices in the dataset.
-        - param custom: {display: column} mapping of additional (text) tooltips.
+        - label: whether to expect and show a "label" field.
+        - text: whether to expect and show a "text" field.
+        - image: whether to expect and show an "image" (url/path) field.
+        - audio: whether to expect and show an "audio" (url/path) field.
+        - coords: whether to show xy-coordinates.
+        - index: whether to show indices in the dataset.
+        - custom: {display: column} mapping of additional (text) tooltips.
     """
     # initialize mutable default value
     custom = custom or dict()
@@ -222,3 +237,54 @@ def bokeh_hover_tooltip(
     divbox += divbox_suffix
     script += script_suffix
     return divbox + script
+
+
+def binder_proxy_app_url(app_path, port=5006):
+    """
+    ???+ note "Find the URL of Bokeh server app in the current Binder session."
+
+        Intended for visiting a Binder-hosted Bokeh server app.
+
+        Will NOT work outside of Binder.
+    """
+
+    service_url_path = os.environ.get(
+        "JUPYTERHUB_SERVICE_PREFIX", "/user/hover-binder/"
+    )
+    proxy_url_path = f"proxy/{port}/{app_path}"
+
+    base_url = "https://hub.gke2.mybinder.org"
+    user_url_path = urljoin(service_url_path, proxy_url_path)
+    full_url = urljoin(base_url, user_url_path)
+    return full_url
+
+
+def remote_jupyter_proxy_url(port):
+    """
+    ???+ note "Callable to configure Bokeh's show method when using a proxy (JupyterHub)."
+
+        Intended for rendering a in-notebook Bokeh app.
+
+        Usage:
+
+        ```python
+        # show(plot)
+        show(plot, notebook_url=remote_jupyter_proxy_url)
+        ```
+    """
+
+    # find JupyterHub base (external) url, default to Binder
+    base_url = os.environ.get("JUPYTERHUB_BASE_URL", "https://hub.gke2.mybinder.org")
+    host = urlparse(base_url).netloc
+
+    if port is None:
+        return host
+
+    service_url_path = os.environ.get(
+        "JUPYTERHUB_SERVICE_PREFIX", "/user/hover-binder/"
+    )
+    proxy_url_path = f"proxy/{port}"
+
+    user_url = urljoin(base_url, service_url_path)
+    full_url = urljoin(user_url, proxy_url_path)
+    return full_url
